;;
;; chicken-sdl2: CHICKEN Scheme bindings to Simple DirectMedia Layer 2
;;
;; Copyright © 2013–2021  John Croisant.
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; - Redistributions of source code must retain the above copyright
;;   notice, this list of conditions and the following disclaimer.
;;
;; - Redistributions in binary form must reproduce the above copyright
;;   notice, this list of conditions and the following disclaimer in
;;   the documentation and/or other materials provided with the
;;   distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;; FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
;; COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
;; INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
;; (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
;; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
;; STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
;; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
;; OF THE POSSIBILITY OF SUCH DAMAGE.


(export create-texture
        create-texture*
        create-texture-from-surface
        create-texture-from-surface*
        destroy-texture!

        query-texture
        query-texture-raw
        texture-format
        texture-access
        texture-w
        texture-h

        lock-texture-raw!
        unlock-texture!
        update-texture-raw!
        update-yuv-texture-raw! ;; SDL >= 2.0.1

        texture-alpha-mod   texture-alpha-mod-set!
        texture-blend-mode  texture-blend-mode-set!
        texture-color-mod   texture-color-mod-set!
        texture-colour-mod  texture-colour-mod-set!
        )


(: create-texture
   (sdl2:renderer* enum enum integer integer
    -> sdl2:texture))
(define (create-texture renderer format access w h)
  (set-finalizer! (create-texture* renderer format access w h)
                  destroy-texture!))

(: create-texture*
   (sdl2:renderer* enum enum integer integer
    -> sdl2:texture))
(define (create-texture* renderer format access w h)
  (let ((format-int (if (integer? format)
                        format
                        (symbol->pixel-format-enum format)))
        (access-int (if (integer? access)
                        access
                        (symbol->texture-access access))))
    (try-call (SDL_CreateTexture renderer format-int access-int w h)
              fail?: (%struct-fail texture?))))


(: create-texture-from-surface
   (sdl2:renderer* sdl2:surface* -> sdl2:texture))
(define (create-texture-from-surface renderer surface)
  (set-finalizer! (create-texture-from-surface* renderer surface)
                  destroy-texture!))

(: create-texture-from-surface*
   (sdl2:renderer* sdl2:surface* -> sdl2:texture))
(define (create-texture-from-surface* renderer surface)
  (try-call (SDL_CreateTextureFromSurface renderer surface)
            fail?: (%struct-fail texture?)))


(: destroy-texture!
   (sdl2:texture* -> void))
(define (destroy-texture! texture)
  (unless (struct-null? texture)
    (SDL_DestroyTexture texture)))


(: query-texture
   (sdl2:texture* -> symbol symbol integer integer))
(define (query-texture texture)
  (receive (format-int access-int w h) (query-texture-raw texture)
    (values (pixel-format-enum->symbol format-int)
            (texture-access->symbol access-int)
            w
            h)))

(: query-texture-raw
   (sdl2:texture* -> integer integer integer integer))
(define (query-texture-raw texture)
  (with-temp-mem ((format-out (%allocate-Uint32))
                  (access-out (%allocate-Sint32))
                  (w-out      (%allocate-Sint32))
                  (h-out      (%allocate-Sint32)))
    (try-call (SDL_QueryTexture texture
                                format-out access-out
                                w-out h-out)
              on-fail: (begin (free format-out)
                              (free access-out)
                              (free w-out)
                              (free h-out)))
    (values (pointer-u32-ref format-out)
            (pointer-s32-ref access-out)
            (pointer-s32-ref w-out)
            (pointer-s32-ref h-out))))


(: texture-format
   (sdl2:texture* -> symbol))
(define (texture-format texture)
  (with-temp-mem ((format-out (%allocate-Sint32)))
    (try-call (SDL_QueryTexture texture format-out #f #f #f)
              on-fail: (free format-out))
    (pixel-format-enum->symbol (pointer-s32-ref format-out))))

(: texture-access
   (sdl2:texture* -> symbol))
(define (texture-access texture)
  (with-temp-mem ((access-out (%allocate-Sint32)))
    (try-call (SDL_QueryTexture texture #f access-out #f #f)
              on-fail: (free access-out))
    (texture-access->symbol (pointer-s32-ref access-out))))

(: texture-w
   (sdl2:texture* -> integer))
(define (texture-w texture)
  (with-temp-mem ((w-out (%allocate-Sint32)))
    (try-call (SDL_QueryTexture texture #f #f w-out #f)
              on-fail: (free w-out))
    (pointer-s32-ref w-out)))

(: texture-h
   (sdl2:texture* -> integer))
(define (texture-h texture)
  (with-temp-mem ((h-out (%allocate-Sint32)))
    (try-call (SDL_QueryTexture texture #f #f #f h-out)
              on-fail: (free h-out))
    (pointer-s32-ref h-out)))


(: lock-texture-raw!
   (sdl2:texture* (or sdl2:rect* boolean) -> pointer integer))
(define (lock-texture-raw! texture rect)
  (with-temp-mem ((pitch-out (%allocate-Sint32)))
    (let ((pixels** (make-pointer-vector 1)))
      (try-call (SDL_LockTexture texture rect pixels** pitch-out)
                on-fail: (free pitch-out))
      (values (pointer-vector-ref pixels** 0)
              (pointer-s32-ref pitch-out)))))

(: unlock-texture!
   (sdl2:texture* -> void))
(define (unlock-texture! texture)
  (SDL_UnlockTexture texture))


(: update-texture-raw!
   (sdl2:texture*
    (or sdl2:rect* false)
    (or pointer locative)
    fixnum
    -> void))
(define (update-texture-raw! texture rect pixels pitch)
  (try-call (SDL_UpdateTexture texture rect pixels pitch))
  (void))


#+libSDL-2.0.1+
(: update-yuv-texture-raw!
   (sdl2:texture*
    (or sdl2:rect* boolean)
    (or pointer locative) integer
    (or pointer locative) integer
    (or pointer locative) integer
    -> void))
(define-versioned (update-yuv-texture-raw!
                   texture rect
                   y-plane y-pitch
                   u-plane u-pitch
                   v-plane v-pitch)
    libSDL-2.0.1+
  (try-call (SDL_UpdateYUVTexture
             texture rect
             y-plane y-pitch
             u-plane u-pitch
             v-plane v-pitch))
  (void))


(: texture-alpha-mod
   (sdl2:texture* -> fixnum))
(define (texture-alpha-mod texture)
  (with-temp-mem ((alpha-out (%allocate-Uint8)))
    (try-call (SDL_GetTextureAlphaMod texture alpha-out)
              on-fail: (free alpha-out))
    (pointer-u8-ref alpha-out)))

(: texture-alpha-mod-set!
   (sdl2:texture* fixnum -> void))
(define (texture-alpha-mod-set! texture alpha)
  (try-call (SDL_SetTextureAlphaMod texture alpha))
  (void))

(set! (setter texture-alpha-mod) texture-alpha-mod-set!)


(: texture-blend-mode
   (sdl2:texture* -> (or symbol integer)))
(define (texture-blend-mode texture)
  (blend-mode->symbol (texture-blend-mode-raw texture)
                      identity))

(: texture-blend-mode-raw
   (sdl2:texture* -> fixnum))
(define (texture-blend-mode-raw texture)
  (with-temp-mem ((mode-out (%allocate-Uint8)))
    (try-call (SDL_GetTextureBlendMode texture mode-out)
              on-fail: (free mode-out))
    (pointer-u8-ref mode-out)))

(: texture-blend-mode-set!
   (sdl2:texture* enum -> void))
(define (texture-blend-mode-set! texture mode)
  (define (bad-mode-err x)
    (error 'texture-blend-mode-set!
           "Invalid texture blend mode" x))
  (let ((mode-int (if (integer? mode)
                      mode
                      (symbol->blend-mode mode bad-mode-err))))
    (try-call (SDL_SetTextureBlendMode texture mode-int))
    (void)))

(set! (setter texture-blend-mode) texture-blend-mode-set!)
(set! (setter texture-blend-mode-raw) texture-blend-mode-set!)


(: texture-color-mod
   (sdl2:texture* -> fixnum fixnum fixnum))
(define (texture-color-mod texture)
  (with-temp-mem ((r-out (%allocate-Uint8))
                  (g-out (%allocate-Uint8))
                  (b-out (%allocate-Uint8)))
    (try-call (SDL_GetTextureColorMod texture r-out g-out b-out)
              on-fail: (begin (free r-out)
                              (free g-out)
                              (free b-out)))
    (values (pointer-u8-ref r-out)
            (pointer-u8-ref g-out)
            (pointer-u8-ref b-out))))

(: texture-color-mod-set!
   (sdl2:texture*
    (or sdl2:color
        (list fixnum fixnum fixnum)
        (list fixnum fixnum fixnum fixnum))
    -> void))
(define (texture-color-mod-set! texture rgb-or-color)
  (receive (r g b) (if (list? rgb-or-color)
                       (values (list-ref rgb-or-color 0)
                               (list-ref rgb-or-color 1)
                               (list-ref rgb-or-color 2))
                       (values (color-r rgb-or-color)
                               (color-g rgb-or-color)
                               (color-b rgb-or-color)))
    (try-call (SDL_SetTextureColorMod texture r g b))
    (void)))

(set! (setter texture-color-mod) texture-color-mod-set!)

(define texture-colour-mod-set! texture-color-mod-set!)
(define texture-colour-mod      texture-color-mod)
